# -*- coding: utf-8 -*-

# Copyright © 2018 by IBPort. All rights reserved.
# @Author: Neal Wong
# @Email: ibprnd@gmail.com

import os
import sys
import logging
import logging.handlers
from datetime import datetime

import click
from pydispatch.robust import sendRobust
from pydispatch import dispatcher
import requests
from bigcommerce.exception import ClientRequestException

import bigcommerce_v3
from bigcommerce_v3.resources.products import ProductVariants
from dropshipping.utils.misc import load_object
from dropbox_utils import DropboxHelper
from cmutils.config_loaders import YamlConfigLoader
from cmutils.stats_collector import StatsCollector
from cmutils.process_checker import ProcessChecker
from cmutils.heart_beater import HeartBeater
from website_utils.price_loaders import BasePriceLoader
from website_utils.exchange_rate import FixerExchangeRate
from website_utils.platform_syncers import PLATFORM_MAPPING, PlatformSyncer
from website_utils.signals import website_started, website_updated, website_finished

from bigcommerce_repricing.price_finders import AliexpressPriceFinder
from bigcommerce_repricing.aliexpress_inventory_manager import AliexpressInventoryManager
from bigcommerce_repricing.utils import format_ali_sku
from bigcommerce_repricing import logger


@click.command()
@click.option('-c', '--config_path', required=True, help='Configuration file path.')
def bigcommerce_aliexpress_reprice(config_path):
    config_path = os.path.abspath(os.path.expanduser(config_path))
    if not os.path.isfile(config_path):
        logger.error('Could not find configuration file %s', config_path)
        sys.exit(1)

    cl = YamlConfigLoader(config_path)
    config = cl.load()

    work_dir = os.path.join(
        os.path.expanduser('~'), '.bigcommerce_reprice', config['store']['name'])
    if not os.path.isdir(work_dir):
        os.makedirs(work_dir)
    config['work_dir'] = work_dir

    heart_beater = HeartBeater(os.path.join(work_dir, '_bp'), 30)

    pc = ProcessChecker(work_dir)
    pids = pc.get_pids()
    pids_count = len(pids)
    if pids_count <= 0:
        pid = -1
        logger.info('No bigcommerce reprice instance is running, start a new one')
    elif pids_count > 1 and heart_beater.is_alive():
        logger.warning('Multiple bigcommerce reprice instances is running, shutdown \
            all instances and start a new one')
        for pid in pids:
            pc.kill_proc(pid)
            pc.remove_pid(pid)

        pid = -1
    else:
        pid = pids[0]

    if pid == -1:
        # Existing repricing process already killed, or no instance is running, do nothing
        pass
    elif pc.is_running(pid):
        logger.info('One bigcommerce reprice instance is running, Exit!')
        sys.exit(0)
    else:
        logger.warning(
            'Current bigcommerce reprice instance is zombie, '
            'shutdown it and start a new one')
        pc.kill_proc(pid, True)
        pc.remove_pid(pid)

    # Save new repricing process information
    pc.save_pid(os.getpid())

    log_dir = os.path.join(work_dir, 'logs')
    if not os.path.isdir(log_dir):
        os.makedirs(log_dir)
    log_path = os.path.join(log_dir, 'bigcommerce_aliexpress_reprice.log')
    level = logging.INFO
    max_bytes = 20 * 1024 ** 2
    fh = logging.handlers.RotatingFileHandler(
        log_path, maxBytes=max_bytes, backupCount=5)
    fh.setLevel(level)
    formatter = logging.Formatter('%(asctime)s %(name)s [%(levelname)s]:%(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    def on_website_updated(records):
        heart_beater.beat()

    core = BigcommerceAliexpressRepriceCore(config)
    dispatcher.connect(on_website_updated, signal=website_updated, sender=core)
    core.run()

    heart_beater.stop()
    pc.remove_pid(os.getpid())


class BigcommerceAliexpressRepriceCore(object):
    def __init__(self, cfg):
        # Setup base attributes, other utilities will use these attributes
        self.cfg = cfg
        self.bigcommerce_api = bigcommerce_v3.api.BigcommerceApi(
            client_id=self.cfg['store']['client_id'],
            store_hash=self.cfg['store']['store_hash'],
            access_token=self.cfg['store']['access_token'])
        fixer_api_key = self.cfg['fixer']['api_key'] if 'fixer' in cfg else None
        self.exchange_rate = FixerExchangeRate(fixer_api_key)
        self.stats = StatsCollector()
        self.finished = False
        self.started_at = None
        self.ended_at = None

        # Only USD can be base currency now
        base_currency = 'USD'
        source_currency = 'USD'

        store = self.bigcommerce_api.Store.all()
        self.currency = store.currency
        if self.currency == source_currency and self.currency == base_currency:
            rates = dict()
            rates[self.currency] = 1
        else:
            rates = self.exchange_rate.get_exchange_rate(base_currency)

        # Setup utilities
        self.inventory_manager = AliexpressInventoryManager(self.bigcommerce_api)
        self.price_loader = BasePriceLoader(self.currency, rates, self.cfg['pricing'])

        self.price_finder = AliexpressPriceFinder(
            self.cfg['offer_service']['host'], self.cfg['offer_service']['port'],
            self.cfg['offer_service']['user'], self.cfg['offer_service']['password'])
        self.batch_size = 500
        self.records_buf = []

        if 'dropbox' in self.cfg and self.cfg['dropbox']['access_token']:
            self.dpx_helper = DropboxHelper(self.cfg['dropbox']['access_token'])
        else:
            self.dpx_helper = None

        inv_dir = os.path.join(self.cfg['work_dir'], 'inventory_report')
        if not os.path.isdir(inv_dir):
            os.makedirs(inv_dir)
        self.inv_path = os.path.join(inv_dir, 'inv.txt')
        if os.path.isfile(self.inv_path):
            os.remove(self.inv_path)
        self.dpx_inv_path = '/'.join([
            '', 'offer_updater', 'aliexpress', self.cfg['store']['name'], 'inv.txt'])

        self.platforms = []
        if 'platforms' in cfg:
            for pname, pconfig in cfg['platforms'].items():
                pconfig['work_dir'] = self.cfg['work_dir']
                pclass_name = PLATFORM_MAPPING.get(pname, None)
                if pclass_name is None:
                    logger.warning('Unsupported platform - %s', pname)
                    continue
                pclass = load_object(pclass_name)
                if not issubclass(pclass, PlatformSyncer):
                    logger.warning('Platform must be a subclass of PlatformSyncer - %s.', pname)
                    continue

                try:
                    platform = pclass(pconfig)
                    self.platforms.append(platform)
                except Exception as e:
                    logger.warning('Failed to initialize platform - %s', pname)
                    logger.exception(e)

    def run(self):
        self.started_at = datetime.now()
        sendRobust(signal=website_started, sender=self)

        try:
            for record in self.inventory_manager.enumerate_inventory(self.inv_path):
                self.records_buf.append(record)
                if len(self.records_buf) < self.batch_size:
                    continue

                records = self.records_buf
                self.update(records)
                self.records_buf = []

            if self.records_buf:
                self.update(self.records_buf)

            self.finished = True
        except (KeyboardInterrupt, SystemExit):
            logger.info('Received keyboard interrupt, exiting...')
        except Exception as e:
            logger.exception(e)
        finally:
            self.ended_at = datetime.now()

            # Trigger finished event
            sendRobust(signal=website_finished, sender=self)

            self.on_website_finished()

    def update(self, records):
        d = dict()

        product_ids = [r['ali_product_id'] for r in records]
        offers = self.price_finder.find_offer_for_products(product_ids)
        if isinstance(offers, bool) and not offers:
            raise RuntimeError('Offer service is unavailable!')

        if offers is None:
            return

        website_records = []
        for r in records:
            offer = offers.get(r['ali_product_id'], d)

            product_id = r['product_id']
            for variant_info in r['variants']:
                self.stats.inc_value('website_processed')

                variant_info['product_id'] = product_id
                variant_id = variant_info['variant_id']
                sku = variant_info['sku']
                price = round(float(variant_info['price']), 2)
                quantity = int(variant_info['quantity'])
                website_record = {
                    'platform': 'bigcommerce',
                    'region': 'US',
                    'product_id': product_id,
                    'variant_id': variant_id
                }

                sku_parts = sku.replace('-', '_').split('_')
                if len(sku_parts) <= 1:
                    offer_key = sku_parts.pop()
                else:
                    sku_attr = sku_parts.pop()
                    offer_key = format_ali_sku(sku_attr)

                source_offer = offer.get(offer_key, None)
                sale_offer = self.price_loader.calc_offer(source_offer)
                if not sale_offer:
                    self.stats.inc_value('website_unchanged')
                    if quantity > 0:
                        self.stats.inc_value('website_in_stock')
                    else:
                        self.stats.inc_value('website_out_of_stock')

                    website_record['price'] = price
                    website_record['quantity'] = quantity
                    website_record['currency'] = self.currency
                    website_records.append(website_record)

                    logger.info(
                        '[SourceOfferNotFound] Skip. variant: %s', repr(variant_info))
                    continue

                variant = ProductVariants(
                    {'product_id': product_id, 'id': variant_id},
                    _connection=self.bigcommerce_api.connection)

                sale_price = sale_offer.get('price', 0)
                sale_quantity = sale_offer.get('quantity', 0)
                website_record['currency'] = offer.get('currency', self.currency)
                if sale_price <= 0 or sale_quantity <= 0:
                    # Out of stock
                    if quantity > 0:
                        try:
                            variant.update(inventory_level=0)
                            quantity = 0

                            website_record.update({
                                'price': price,
                                'quantity': quantity
                            })

                            self.stats.inc_value('website_changed')
                            logger.info(
                                '[OfferUpdated] source_offer: %s, offer: %s, variant: %s',
                                source_offer, repr(sale_offer), repr(variant_info))
                        except ClientRequestException as e:
                            logger.exception(e)

                            website_record.update({
                                'price': price,
                                'quantity': quantity
                            })

                            self.stats.inc_value('website_unchanged')
                            logger.warning(
                                '[OfferUpdateFailed] source_offer: %s, offer: %s, variant: %s',
                                source_offer, repr(sale_offer), repr(variant_info))
                    else:
                        self.stats.inc_value('website_unchanged')

                        logger.info(
                            '[OfferUnchanged] variant: %s', repr(variant_info))

                    self.stats.inc_value('website_out_of_stock')
                else:
                    if ((sale_price - price) <= self.cfg['pricing']['max_allowed_price_change'] and \
                        (sale_quantity > 0 and quantity > 0)) or \
                        (sale_quantity <= 0 and quantity <= 0):
                        # No need to update
                        self.stats.inc_value('website_unchanged')
                        if quantity > 0:
                            self.stats.inc_value('website_in_stock')
                        else:
                            self.stats.inc_value('website_out_of_stock')

                        website_record.update({
                            'price': price,
                            'quantity': quantity
                        })

                        logger.info(
                            '[OfferUnchanged] source_offer: %s, offer: %s, variant: %s',
                            repr(source_offer), repr(sale_offer), repr(variant_info))
                        continue

                    try:
                        variant.update(price=sale_price, inventory_level=sale_quantity)

                        self.stats.inc_value('website_changed')

                        website_record.update({
                            'price': sale_price,
                            'quantity': sale_quantity
                        })

                        variant_info['price'] = sale_price
                        variant_info['quantity'] = sale_quantity
                        logger.info(
                            '[OfferUpdated] source_offer: %s, offer: %s, variant: %s',
                            repr(source_offer), repr(sale_offer), repr(variant_info))
                    except ClientRequestException as e:
                        logger.exception(e)

                        website_record.update({
                            'price': price,
                            'quantity': quantity
                        })

                        self.stats.inc_value('website_unchanged')
                        logger.warning(
                            '[OfferUpdateFailed] source_offer: %s, offer: %s, variant: %s',
                            source_offer, repr(sale_offer), repr(variant_info))

                    if website_record['quantity'] > 0:
                        self.stats.inc_value('website_in_stock')
                    else:
                        self.stats.inc_value('website_out_of_stock')

                website_records.append(website_record)

        # Trigger website updated event
        sendRobust(signal=website_updated, sender=self, records=website_records)

    def on_website_finished(self):
        time_format = '%Y-%m-%dT%H:%M:%S'

        if self.dpx_helper and self.finished:
            try:
                self.dpx_helper.files_upload(self.inv_path, self.dpx_inv_path)
            except Exception as e:
                logger.exception(e)

        if 'report' in self.cfg:
            payload = {
                'name': self.cfg['store']['name'],
                'started_at': self.started_at.strftime(time_format),
                'ended_at': self.ended_at.strftime(time_format),
                'inventory_total': self.stats.get_value('inventory_total', 0),
                'inventory_processed': self.stats.get_value('inventory_processed', 0),
                'inventory_invalid': self.stats.get_value('inventory_invalid', 0),
                'inventory_skipped': self.stats.get_value('inventory_skipped', 0),
                'website_processed': self.stats.get_value('website_processed', 0),
                'website_changed': self.stats.get_value('website_changed', 0),
                'website_unchanged': self.stats.get_value('website_unchanged', 0),
                'website_in_stock': self.stats.get_value('website_in_stock', 0),
                'website_out_of_stock': self.stats.get_value('website_out_of_stock', 0)
            }
            for platform in self.platforms:
                payload.update(platform.get_stats())
            try:
                resp = requests.post(
                    self.cfg['report']['endpoint'], json=payload,
                    auth=(self.cfg['report']['user'], self.cfg['report']['password']))
                logger.info('[Report] payload: %s, response: %s', repr(payload), repr(resp.json()))
            except Exception as e:
                logger.exception(e)

        time_used = round((self.ended_at - self.started_at).total_seconds() / 3600, 2)
        logger.info(
            '[BigcommerceAliexpressRepricerFinished] '
            'started_at: %s, ended_at: %s, time_used: %.2fhours',
            self.started_at.strftime(time_format), self.ended_at.strftime(time_format),
            time_used)

if __name__ == '__main__':
    bigcommerce_aliexpress_reprice()
