# -*- coding: utf-8 -*-

# Copyright © 2019 by IBPort. All rights reserved.
# @Author: Neal Wong
# @Email: ibprnd@gmail.com

import os
import sys
import logging
import logging.handlers
from datetime import datetime

import click
import requests

from cmutils.config_loaders import YamlConfigLoader
from dropbox_utils.dropbox_helper import DropboxHelper
from website_utils.google_shopping import GoogleShopping
from website_utils import logger


@click.command()
@click.option('-c', '--config_path', required=True, help='Configuration file path.')
def cleanup_google_merchant(config_path):
    config_path = os.path.abspath(os.path.expanduser(config_path))
    if not os.path.isfile(config_path):
        logger.error('Could not find configuration file %s', config_path)
        sys.exit(1)

    cl = YamlConfigLoader(config_path)
    cfg = cl.load()

    work_dir = os.path.join(
        os.path.expanduser('~'), '.bigcommerce_reprice', cfg['store']['name'])
    if not os.path.isdir(work_dir):
        os.makedirs(work_dir)

    log_dir = os.path.join(work_dir, 'logs')
    if not os.path.isdir(log_dir):
        os.makedirs(log_dir)
    log_path = os.path.join(log_dir, 'google_merchant_cleanup.log')
    level = logging.INFO
    max_bytes = 20 * 1024 ** 2
    fh = logging.handlers.RotatingFileHandler(
        log_path, maxBytes=max_bytes, backupCount=5)
    fh.setLevel(level)
    formatter = logging.Formatter('%(asctime)s %(name)s [%(levelname)s]:%(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    if 'google_shopping' not in cfg['platforms']:
        logger.error('Missing required configuration "platforms>google_shopping"')
        sys.exit(1)

    google_shopping = GoogleShopping(cfg['platforms']['google_shopping'])
    issues_to_ignore = cfg['platforms']['google_shopping'].get(
        'issues_to_ignore', ['image_link_pending_crawl'])
    deleted_products, spam_products = google_shopping.cleanup(issues_to_ignore)

    if 'dropbox' not in cfg or not cfg['dropbox']['access_token']:
        return
    dpx_helper = DropboxHelper(cfg['dropbox']['access_token'])
    if deleted_products:
        local_deleted_save_path = os.path.join(work_dir, 'platforms', 'google_shopping', 'deleted_products.txt')

        deleted_prods = dict()
        if os.path.isfile(local_deleted_save_path):
            with open(local_deleted_save_path) as fh:
                for line in fh:
                    record = line.strip()
                    mid, pid, _ = record.split('\t')
                    deleted_prods['{}#{}'.format(mid, pid)] = record

        for prod in deleted_products:
            mid, pid = prod.split('\t')[:2]
            k = '{}#{}'.format(mid, pid)
            if k in deleted_prods:
                continue

            deleted_prods[k] = prod

        with open(local_deleted_save_path, 'w') as fh:
            fh.write('\n'.join(deleted_prods.values()))

        dpx_deleted_path = '/'.join(
            ['', 'website_reprice', cfg['store']['name'], 'gm_deleted.txt'])
        dpx_helper.files_upload(local_deleted_save_path, dpx_deleted_path)

    if spam_products:
        local_spam_save_path = os.path.join(work_dir, 'platforms', 'google_shopping', 'spam_products.txt')

        spam_prods = dict()
        if os.path.isfile(local_spam_save_path):
            with open(local_spam_save_path) as fh:
                for line in fh:
                    record = line.strip()
                    mid, pid, _ = record.split('\t')
                    spam_prods['{}#{}'.format(mid, pid)] = record

        for prod in spam_products:
            mid, pid = prod.split('\t')[:2]
            k = '{}#{}'.format(mid, pid)
            if k in deleted_prods:
                continue

            spam_prods[k] = prod

        with open(local_spam_save_path, 'w') as fh:
            fh.write('\n'.join(spam_prods.values()))

        dpx_spam_path = '/'.join(
            ['', 'website_reprice', cfg['store']['name'], 'gm_spam.txt'])
        dpx_helper.files_upload(local_spam_save_path, dpx_spam_path)


if __name__ == '__main__':
    cleanup_google_merchant()
