# -*- coding: utf-8 -*-

# Copyright © 2018 by IBPort. All rights reserved.
# @Author: Neal Wong
# @Email: ibprnd@gmail.com

import json
from datetime import datetime

from dropshipping.utils.offer_filters import LowestOfferListingOfferFilter
from dropshipping.utils.offer_services import EsLowestOfferListingOfferService
from dropshipping.utils.offer_converters import (
    EsLowestOfferListingOfferConverter,
    MwsLowestOfferListingOfferConverter
)
from dropshipping.utils.price_finders import (
    OfferServiceLowestOfferListingPriceFinder,
    MwsLowestOfferListingPriceFinder
)
from dropshipping.utils.es_service import EsService
from dropshipping.mws import Products
from website_utils.price_finders import MixedPriceFinder
from website_utils.legacy import (
    LegacyOfferServiceLowestOfferListingPriceFinder,
    LegacyEsOfferService
)

from bigcommerce_repricing import legacy_support
from bigcommerce_repricing import logger
from bigcommerce_repricing.utils import format_ali_sku


class AmzPriceFinder(object):
    def __init__(self, cfg):
        self.source_marketplace = cfg['store']['marketplace'].upper()
        self.price_finder = MixedPriceFinder()
        self.price_finder.offer_alive_hours = cfg['store'].get('offer_alive_hours', 48)

        self.offer_filter = LowestOfferListingOfferFilter({
            'feedback': cfg['pricing.provider'].get('min_seller_feedback', 100),
            'rating': cfg['pricing.provider'].get('min_seller_rate', 90),
            'domestic': cfg['pricing.provider'].get('ships_domestically', True),
            'shipping_time': cfg['pricing.provider'].get('max_shipping_days', 4),
            'subcondition': cfg['pricing.provider'].get('subcondition', 70),
            'fba': cfg['pricing.provider'].get('fba', None)
        })

        if 'offer_service' in cfg and cfg['offer_service']:
            if legacy_support:
                self.offer_service = LegacyEsOfferService(
                    cfg['offer_service']['host'], cfg['offer_service']['port'],
                    cfg['offer_service']['user'], cfg['offer_service']['password'])
                self.offer_service_price_finder = LegacyOfferServiceLowestOfferListingPriceFinder(
                    None, self.offer_service, None)
            else:
                self.offer_service = EsLowestOfferListingOfferService(
                    cfg['offer_service']['host'], cfg['offer_service']['port'],
                    cfg['offer_service']['user'], cfg['offer_service']['password'])
                offer_converter = EsLowestOfferListingOfferConverter()
                self.offer_service_price_finder = OfferServiceLowestOfferListingPriceFinder(
                    self.offer_filter, self.offer_service, offer_converter)

            self.price_finder.add_price_finder(self.offer_service_price_finder)
            self.buf_size = 500
        else:
            self.offer_service_price_finder = None

        if 'mws' in cfg and cfg['mws']:
            access_key = cfg['mws']['aws_access_key_id']
            secret_key = cfg['mws']['aws_secret_access_key']
            seller_id = cfg['mws']['seller_id']
            auth_token = cfg['mws'].get('mws_auth_token', None)
            marketplace = self.source_marketplace
            self.products_api = Products(
                access_key, secret_key, seller_id, marketplace, auth_token=auth_token)
            offer_converter = MwsLowestOfferListingOfferConverter()
            self.mws_price_finder = MwsLowestOfferListingPriceFinder(
                self.offer_filter, self.products_api, offer_converter)
            self.price_finder.add_price_finder(self.mws_price_finder)
            self.buf_size = 20
        else:
            self.mws_price_finder = None

        if self.mws_price_finder is None and self.offer_service_price_finder is None:
            raise ValueError('No price finder configuration is available!')

    def find_offer_for_asins(self, asins, country, condition, excludeme=True):
        return self.price_finder.find_offer_for_asins(asins, country, condition, excludeme)


class AliexpressPriceFinder(EsService):
    def find_offer_for_products(self, product_ids):
        """Search available offer on server.

        Returns:
          None - Not found on elastic search server
          False - Elastic search unavailable or could not connect to elastic search
          dict - Result on elastic search
        """

        if not self.is_active():
            return False

        product_ids_count = len(product_ids)
        params = {
            'index': 'aliexpress',
            'doc_type': '_doc',
            'size': product_ids_count,
            'body': {
                'query': {'terms': {'_id': product_ids}}
            }
        }

        resp = self._search(**params)
        if not resp:
            return resp

        result = dict()
        for item in resp['hits']['hits']:
            offer = dict()

            record = item['_source']
            for sku_product in json.loads(record['sku_products']):
                if 'skuAttr' in sku_product:
                    item_key = format_ali_sku(sku_product['skuAttr'])
                else:
                    item_key = record['product_id']

                offer[item_key] = {
                    'price': self._get_sku_price(sku_product['skuVal']),
                    'quantity': self._calc_sku_quantity(sku_product['skuVal'])
                }

            if 'time' in record:
                try:
                    offer['time'] = datetime.strptime(record['time'][:19], '%Y-%m-%dT%H:%M:%S')
                except ValueError:
                    pass

            result[item['_id']] = offer

        return result

    def _calc_sku_quantity(self, sku_val):
        return int(str(sku_val.get('availQuantity', sku_val.get('inventory', 0))).replace(',', ''))

    def _get_sku_price(self, sku_val):
        price = 0

        price_prioritize_keys = [
            'actSkuMultiCurrencyDisplayPrice', 'skuMultiCurrencyDisplayPrice',
            'actSkuMultiCurrencyCalPrice', 'skuMultiCurrencyCalPrice',
            'actSkuCalPrice', 'skuCalPrice'
        ]

        for key in price_prioritize_keys:
            price = float(str(sku_val.get(key, 0)).replace(',', ''))
            if price != 0:
                break

        if price == 0:
            price = 100

        return price
